#define TUNES_VERSION "alpha-980907" /* yymmdd */
#define IMAGE_NAME "tunes.img"

#define DEBUG /**/
#define _DEBUG /**/
#define debug printf
/* Notes:

axioms: Copy,move,del
you can copy a container to the root
you can copy axioms to it
You can copy it to the evaluator

you can copy objects to the tty to view them

copy 0 tty
copy object 0
copy 0 tty

the system is saved at exit, and loaded at startup
the special axiom 'quit' does as expected...(possibly only accessible from
the command line)

the command line is interpreted as an object that is evaluated when you
press Enter.  You can evaluate stored expressions by typing the number.  I
recommend not typing any low-numbered object numbers to evaluate.  Typing
0 will evaluate every object (probably not what you want). 

Evaluation works as follows: Objects are descended into.  Sub-objects are
evaluated in numerical order.

The next free object number is always used when objects are created.

spec: every object that is created in the system (not ones that come
prefab) gets a spec.  The spec contains an entry for each property.  For
instance, every object has a property corresponding to its type.  In this
system, each object will also have a property corresponding to its parent
object.  Special types (ones that are prefab) will have a
special entry in the spec that refers to the prefab type.

Container type: A container may hold any number of any type of object.
ok maybe we'll limit the number...16?? for now.
Containers are just lists of object numbers..
because the objects listed are said to be "in" the container.

Don't put a container in itself and then try to evaluate it, because it
will get in an endless loop.

Image file format:  Since every object has a spec, the image file can
simply list the spec for all the objects. (All objects are accessible
within the system by going to the list of specs)  This should be an object
number, followed by its value.  Containers are ordered lists (16 long, to
start with)..  The entries in the list are always an object number.

0. object= Array of int
1. dir= Array of container structures
2. eval
3. tty
4. copy= Syntax: 'copy' 'from' 'to' (from and to are object numbers)
5. move= ditto
6. del= Syntax: 'del' 'object number'
7. quit

implementation...
if the image file doesn't exist, the above 7 objects will be created...

What happens? (no arguments)
0-6. Undefined
7. Exit environment
8+. Evaluate object

(2 args)
0-5. Undefined
6. Second arg:
  0-7. Undefined
  8+. Delete object
7. Undefined
8+. Evaluate first arg's contents, then second arg's contents

(3 args)
0-3. Undefined
4. Second arg:
  0. Third arg:
    0. Undefined
    1. Create new object in root
    2. Do nothing
    3. Display "internal #0: Object"
    4-7. Undefined
    8+. Create new object here
  1. Third arg:
    0-2. Undefined
    3. Display object directory
    4-7. Undefined
    8+. Create dir token here
  2. Third arg:
    0-2. Undefined
    3. Display "internal #2: Evaluator"
    4-7. Undefined
    8+. Create evaluator token here
  3. Third arg:
    0-2. Undefined
    3. Display "internal #3: TTY"
    4-7. Undefined
    8+. Create tty token here
  4. Third arg:
    0-2. Undefined
    3. Display "internal #4: Copy"
    4-7. Undefined
    8+. Create copy token here
  5. Third arg:
    0-2. Undefined
    3. Display "internal #5: Move"
    4-7. Undefined
    8+. Create move token here
  6. Third arg:
    0-2. Undefined
    3. Display "internal #6: Delete"
    4-7. Undefined
    8+. Create delete token here
  7. Third arg:
    0-2. Undefined
    3. Display "internal #7: Quit"
    4-7. Undefined
    8+. Create quit token here
  8+. Third arg:
    0. Undefined
    1. Duplicate object in root
    2. Evaluate object
    3. Display object
    4-7. Undefined
    8+. Duplicate object here
5. Second arg:
  0-7. Undefined
  8+. Third arg:
    0-1. Undefined
    2. Evaluate object and delete
    3. Display object and delete
    4-7. Undefined
6-7. Undefined
8+. Evaluate arg1, then arg2, then arg3
TODO: 4 1-7 1?  (Create alias for primitive)

(4+ Args)
0-7. Undefined
8+. Iteratively evaluate all args.

Programming:

data structures. Object= unsigned short [16]
Object directory= unsigned short**1024

*/

/* internal functions for TUNES commands:
   Argument ranges are checked before calling;
   Syntax errors in object structures are checked within the functions.
*/
/**********
 Prototypes
 **********/
short next_object_number(void);
short *new_obj(void);
short new_top(void); /* 4 0 1 */
short dup_sub(short from, short to); /* 4 1+ 8+ */
short dup_top(short n); /* 4 8+ 1 */
short save(char *); /* called by quit() */
short quit(void); /* 7 */
short del(short n); /* 6 8+ */
short display(short n); /* 4 0+ 3 */
short load(char *); /* read the image file into memory */
short default_primitives(void); /* Create objects 1-7 */
int main(void); /* load, default_primitives, input_loop */
void input_loop(void); /* read commands and do them */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define TUNESERR_DOESNOTEXIST -1
#define TUNESERR_SYNTAXNOTDEFINED -2
#define TUNESERR_ACCESSDENIED -3
#define TUNESERR_DIRFULL -4
#define TUNESERR_NOMEMORY -5
#define TUNESERR_FAILEDTOQUIT -6 /* UNUSED Anymore */
#define TUNESERR_FILEOPENREAD -7
#define TUNESERR_BADIMAGEFILEHEADER -8
#define TUNESERR_BADIMAGEFILEVERSION -9
#define TUNESERR_BADIMAGEFILEOBJDIRSIZE -10
#define TUNESERR_BADIMAGEFILEOBJSIZE -11
#define TUNESERR_BYTECOUNTMISMATCH -12
#define TUNESERR_FILEOPENWRITE -13
#define TUNESERR_OBJECTFULL -14

/* Objects are numbered from 0 to OBJDIRSIZE-1.  You can set this to
anything as long as it is less than 32767 (half of a short).  Objects
in the dir are stored by pointer to type 'tunesobj', which is next. */
#define OBJDIRSIZE 1024
short *dir[OBJDIRSIZE]; 

/* Objects have from 0 to OBJSIZE-1 slots for sub-objects (stored by
number) */
#define OBJSIZE 16
typedef short tunesobj[OBJSIZE];

/* Displaying a primitive causes its name to be printed, except for the
directory (#1), which causes a list of all existing objects to be printed.
*/
char *primitive_names[] = {
"Object", "", "Evaluator", "Tty", "Copy", "Move", "Delete", "Quit"
};

/********************
 Function definitions
 ********************/

/* Create a new pointer to an object in the directory.
   This is just like hard linking in UNIX; neither object is the
   "real" one; the object gets two numbers that both access the same
   contents.  The object remains until all its directory entries are gone.

   Returns: 0=success, TUNESERR_DIRFULL
*/
short dup_top(short n)
{
	short nextfree;
	nextfree=next_object_number();
	if (nextfree==OBJDIRSIZE)
		return TUNESERR_DIRFULL;
	dir[nextfree]=dir[n];
}

/* Create new object in root

Scans for NULL entries in dir[] to store a pointer to a new object.
New object is filled with -1.

Return code: >8=new object number, TUNESERR_DIRFULL, TUNESERR_NOMEMORY
*/
short new_top(void)
{
	short nextfree, *objptr;

	/* find next free dir entry */
	nextfree=next_object_number();
	if (nextfree==OBJDIRSIZE)
		return TUNESERR_DIRFULL;

	/* try to create a new object */
	objptr = new_obj();
	if (!objptr)
		return TUNESERR_NOMEMORY;

	/* assign object to directory */
	dir[nextfree] = objptr;

	/* success */
	return nextfree;
}

/* Create or duplicate new object or token here

Arguments: src can be any short 0..OBJDIRSIZE.  If src is 0, a new object
is created in dest.  If src is any other number, it is put into dest as a
link.

Algorithm: If parent is valid, create a new object and put its number in
parent.

Return code: 0=success, TUNESERR_DOESNOTEXIST (dest does not exist),
TUNESERR_NOMEMORY, TUNESERR_DIRFULL, TUNESERR_OBJECTFULL

Objects will not keep track of how many subobjects, because the subobjects
may not necessarily be in order.  New objects should be placed in the
first 0 slot in their parent.
*/
short dup_sub(short src, short n)
{
	short *destptr, index;
	short err;

	debug("copy from %u to %u\n",src,n);

	/* Get object ref */
	destptr = dir[n];

	/* Is it valid? */
	if (!destptr)
		return TUNESERR_DOESNOTEXIST;

	/* Check for slots */
	for (index=0;index<OBJSIZE;index++)
		if (destptr[index]<0) goto ok;

	return TUNESERR_OBJECTFULL;

ok:
	debug("copy: slot in dest: %u\n", index);

	if (src==0)
	{
		debug("copy: New object\n");

		/* Create new object in directory */
		err = new_top();
		if (err < 0)
			return err;
	}
	
	/* Link to object in specified location */
	if (src==0)
		/* Link to new object number returned from new_top */
		destptr[index]=err;
	else
		/* Link to source */
		destptr[index]=src;

	/* Done! */
	return 0;
}

/* Create new object in memory

Tries to allocate memory for an object and fills it with -1.
Called by new_top.

Return code: NULL=out of memory, other=ptr to object
*/
short *new_obj(void)
{
	short *objptr, loop;

	/* try to allocate memory */
	objptr = malloc(sizeof(tunesobj));

	/* on success, fill object with -1 (any negative number will work)*/
	if (objptr)
		for (loop=0;loop<OBJSIZE;loop++)
			objptr[loop] = -1;

	return objptr;
}

/* Find next available (non-NULL) root directory entry.
   Returns 0..OBJDIRSIZE-1 for success, or OBJDIRSIZE if dir is full.
*/
short next_object_number(void)
{
	short nextfree=0;
	while (nextfree < OBJDIRSIZE && dir[nextfree] != NULL)
		nextfree++;
	return nextfree;
}


/* For the directory (#1), print the ranges of object numbers used.
     Example: 1-10, 13-15, 21-35
   For other primitives, print "Internal #N: <object name>".
   For objects>8, print all the non-negative object numbers separated by
tabs.

   Returns: 0=success, TUNESERR_DOESNOTEXIST
*/
short display(short n)
{
	/* case 1: printing the object directory */
	if (n==1)
	{
		short start, end; /* a range of continguous object numbers */

		/* Algorithm:
			1. find a used object number
			2. find the next non-used object number
			3. print the range
			4. repeat until dir is exhausted */

		/* Note: this loop modifies the loop variable */
		for (start=0;start<OBJDIRSIZE;start++)
		{
			/* Find the next used object number */
			while (start<OBJDIRSIZE && dir[start]==NULL)
				start++;

			if (start==OBJDIRSIZE)
				break;

			/* Find the next unused object number */
			end=start+1;
			while (end<OBJDIRSIZE && dir[end]!=NULL)
				end++;

			if (end==start+1)
				printf("\t%d",start);
			else
				printf("\t%d-%d",start,end-1);

			start=end;
		}

		printf("\n");
	}
	else if (n<8)
		printf("Internal #%u: %s\n",n,primitive_names[n]);
	else
	{
		short index, *objptr;

		debug("display: showing object %u\n",n);

		/* get object reference */
		objptr=dir[n];

		/* check existance */
		if (objptr==NULL)
			return TUNESERR_DOESNOTEXIST;

		/* go through the object */
		for (index=0;index<OBJSIZE;index++)
			/* print all non-negative fields */
			if (objptr[index]>=0)
				printf("\t%d",objptr[index]);

		printf("\n");
	}

	return 0;
}

/* Erase an object.  Don't erase primitives (#0-7).  For any other object,
delete it from the directory.

Returns: 0=success, TUNESERR_ACCESSDENIED=cannot delete primitive,
TUNESERR_DOESNOTEXIST=object does not exist

Future additions: The directory should be able to keep track of all
objects referencing a given object, so when it is deleted, all references
can be informed.  Wait until primitives are reflected to do this, however,
because it needs to be an integral part of a (specialized) set of
primitive copy,move,and delete operations.
*/
short del(short n)
{
	/* Dummy-proof deletion of primitives (for now) */
	if (n<8)
		return TUNESERR_ACCESSDENIED;

	/* Does the object exist in the directory? */
	if (dir[n]==NULL)
		return TUNESERR_DOESNOTEXIST;

	debug("del: Removing object #%u\n",n);

	/* Ignore contents of object.  If any objects refer to this one,
	   too bad. */
	free(dir[n]);

	/* Set this dir entry to available (NULL) */
	dir[n] = NULL;

	return 0; /* success */
}

/* Overwrites the image file.  File format: "TUNESIMAGE", short version=0,
short/OBJDIRSIZE, short/OBJSIZE, 0..OBJDIRSIZE arrays of size OBJSIZE
shorts.  Objects that are not used are filled with the value OBJSIZE,
since this is always an invalid value within an object.

Returns: 0=success, TUNESERR_FILEOPENWRITE, TUNESERR_BYTECOUNTMISMATCH

NOTE: Add a verify() later to verify the current contents of the image
file with the current objects in memory (for use right after saving, as an
extra precaution)

*/
short save(char *image_file)
{
	short out, dirindex, objindex;
	unsigned long total = 0; /* total bytes written */

	FILE *f;

	f=fopen(image_file, "w");
	if (f==NULL)
		return TUNESERR_FILEOPENWRITE;

	total+=fwrite(&"TUNESIMAGE", 10, 1, f);

	out=0; /* Image file version number */
	total+=fwrite(&out, 2, 1, f);

	out=OBJDIRSIZE;
	total+=fwrite(&out, 2, 1, f);

	out=OBJSIZE;
	total+=fwrite(&out, 2, 1, f);

	/* set out to the filler for non-existant objects (any negative #) */
	out = -1;

	for (dirindex=0;dirindex<OBJDIRSIZE;dirindex++)
	{
		if (dir[dirindex]==NULL)
			for (objindex=0;objindex<OBJSIZE;objindex++)
				total+=fwrite(&out, 2, 1, f);
		else
			for (objindex=0;objindex<OBJSIZE;objindex++)
				total+=fwrite(&dir[dirindex][objindex], 
					2, 1, f);
	}

	fclose(f);

	if (total != 16 + OBJDIRSIZE*OBJSIZE)
		return TUNESERR_BYTECOUNTMISMATCH;
	else
		return 0;
}

/* Load image file.

Only handle static objsize/objdirsize (i.e. must have separate executables
of this program to handle image files with different objsize/objdirsize).

Recreates the state of the object directory from the image file.

Returns: 
TUNESERR_FILEOPENREAD: The image file could not be opened for reading
TUNESERR_BADIMAGEFILEHEADER: First 10 bytes weren't "TUNESIMAGE"
TUNESERR_BADIMAGEFILEVERSION: Offset 11 wasn't zero
TUNESERR_BADIMAGEFILEOBJDIRSIZE: dirsize didn't match
TUNESERR_BADIMAGEFILEOBJSIZE: objsize didn't match
TUNESERR_NOMEMORY: out of memory, dir[] is corrupt
 (change this to fill the rest of dir with NULL)
TUNESERR_BYTECOUNTMISMATCH: read the wrong # of bytes from image file
*/
short load(char *image_file)
{
	short in, dirindex, objindex, this_object_is_empty;
	unsigned long total = 0; /* total bytes read */
	char hdr[11];
	short *obj;

	FILE *f;

	f=fopen(image_file, "r");
	if (f==NULL)
		return TUNESERR_FILEOPENREAD;

	/* Only accept image files that begin with "TUNESIMAGE" */
	total+=fread(hdr, 10, 1, f);
	hdr[10]='\0';
	if (strcmp(hdr,"TUNESIMAGE")!=0)
		return TUNESERR_BADIMAGEFILEHEADER;

	/* The version number must be 0 */
	total+=fread(&in, 2, 1, f);
	if (in!=0)
		return TUNESERR_BADIMAGEFILEVERSION;

	/* The objdirsize must be the same as the global constant */
	total+=fread(&in, 2, 1, f);
	if (in!=OBJDIRSIZE)
		return TUNESERR_BADIMAGEFILEOBJDIRSIZE;

	/* The objsize must be the same as the global constant */
	total+=fread(&in, 2, 1, f);
	if (in!=OBJSIZE)
		return TUNESERR_BADIMAGEFILEOBJSIZE;

	for (dirindex=0;dirindex<OBJDIRSIZE;dirindex++)
	{
		/* If this is the first object, or if the last object
		   was added to the directory, allocate memory for
		   a new object. */
		if (obj==NULL)
		{
			obj=malloc(sizeof(tunesobj));
			if (obj==NULL)
				return TUNESERR_NOMEMORY;
		}

		/* Read the object */
		total+=fread(obj, OBJSIZE, 1, f);

		/* Check whether the object is used.  Unused objects
		   were filled with negative numbers by save() */
		this_object_is_empty=1;
		for (objindex=0;objindex<OBJSIZE;objindex++)
			if (obj[objindex]<0)
			{
				this_object_is_empty=0;
				break;
			}

		/* If the object is empty, put NULL in the directory. */
		if (this_object_is_empty)
			dir[dirindex]=NULL;
		/* If it is not empty, put it in the directory and set
		   obj to NULL so it will be reallocated for the next
		   object. */
		else
		{
			dir[dirindex]=obj;
			obj=NULL;
		}
	}

	fclose(f);

	/* Check that the correct number of bytes were read */
	if (total != 16 + OBJDIRSIZE*OBJSIZE)
		return TUNESERR_BYTECOUNTMISMATCH;
	else
		return 0;
}

/* Writes all objects to disk.  If successful, does not return.  If save()
fails, quit() returns in the hope that the objects can be backed up over
the network. It returns the error code from save(). */
short quit(void)
{
	short each;
	short err;

	/* save all the objects to disk */
	err=save(IMAGE_NAME);
	if (err<0)
		return err; /* quit failed */

	/* free all the objects */
	for (each=0;each<OBJDIRSIZE;each++)
		if (dir[each]!=NULL)
			free(dir[each]);

	exit(0);
}

/* Evaluate

Returns: 0=success (side effects depend on expression),
        TUNESERR_DOESNOTEXIST=object does not exist (eval or del),
	TUNESERR_SYNTAXNOTDEFINED=syntax not defined,
	TUNESERR_ACCESSDENIED (delete primitive),
	TUNESERR_DIRFULL, TUNESERR_NOMEMORY (new_top),
	TUNESERR_FILEOPENWRITE or TUNESERR_BYTECOUNTMISMATCH from quit(),
		which passed them on from save()

An object copied to the evaluator is treated as an expression, and
evaluation is attempted.  An expression is either a string of 0-OBJSIZE of
object numbers greater than 7, each of which must refer to a valid
expression, or a string of 1-3 object numbers less than or equal to 7 in
one of the syntaxes defined below:

unary commands (0 arguments):
7) Quit						- quit()

binary commands (1 argument):
6 8+) Delete object				- del()

tertiary commands (2 arguments):
4 0 1) Create new object in root		- new_top()
4 0,2-7 3) Display name of primitive object	- display()
4 0 8+) Create new object here			- new_sub()
4 1 3) Display object directory			- display()
4 1-7 8+) Create token for primitive here	- dup_sub()
4 8+ 1) Duplicate object in root		- dup_top()
4 8+ 2) Evaluate object [same as unary command Evaluate object] - eval()
4 8+ 3) Display object				- display()
4 8+ 8+) Duplicate object here [from, to]	- dup_sub()
5 8+ 2) Evaluate object and delete		- eval(), del()
5 8+ 3) Display object and delete		- display, del()

The other valid expression is the object number 7 (quit), which can be
evaluated directly.

*/
short eval(short n)
{
	short *objptr, index, len, arg[3];
	short err;

	debug("eval: evaluating object #%d\n",n);

	/* evaluate one-arg primitives directly */
	if (n==7)
	{
		return quit(); /* Return quit error code on failure */
	}

	/* get object */
	objptr = dir[n];
	if (!objptr)
		return TUNESERR_DOESNOTEXIST;

	debug ("eval: expression=");

	/* determine length of expression */
	len=0;
	for (index=0;index<OBJSIZE;index++)
	{
		if (objptr[index]!=-1) debug("%d ",objptr[index]);

		if (objptr[index]>=0)
		{
			/* Save first 3 numbers in arg[] */
			if (len<3)
				arg[len]=objptr[index];
			len++;
		}
	}

	debug("(length=%d)\n",len);

	/* Null expression = No operation */
	if (len==0)
	{
		debug("eval: NOOP\n");

		return 0; /* Noop successful */
	}
	else if (arg[0]>7) /* 8+ ... */
	{
		/* Reiterate through obj because there may be >3 exprs */
		for (index=0;index<OBJSIZE;index++)
			if (objptr[index]!=-1)
			{
				err=eval(objptr[index]);
				if (err<0) /* Stop on error */
					return err;
			}
	}
	else if (len==1) /* Evaluate one-arg primitives by calling self */
		return eval(arg[0]);
	else if (len==2)
	{
		/* 6 8+) Delete object */
		if (arg[0]!=6)
			return TUNESERR_SYNTAXNOTDEFINED;
		return del(arg[1]);
	}
	else if (len==3)
	{
	/* Dispatch on arg[0] (copy or move), then arg[2] (destination) */
		if (arg[0]==4)
		{
/* Dispatch on arg[2]:
 X X 0 invalid
int new_top(void);	-- 4 0 1
void dup_top(short n);	-- 4 8+ 1
int eval(short n); 	-- 4 0+ 2
5 8+ 2) Evaluate object and delete
void display(short n);	-- 4 0+ 3
5 8+ 3) Display object and delete
4 0 4-7 invalid
void dup_sub(short from, short to);	-- 4 1+ 8+
*/
			if (arg[2]==0)
				return TUNESERR_SYNTAXNOTDEFINED;
			else if (arg[2]==1)
			{
				if (arg[1]==0) /* 4 0 1 */
					return new_top();
				else /* 4 1+ 1 */
					return dup_top(arg[1]);
			}
			else if (arg[2]==2) /* 4 0+ 2 */
				return eval(arg[1]);
			else if (arg[2]==3) /* 4 0+ 3 */
				return display(arg[1]);
			else if (arg[2]>3 && arg[2]<8)
				return TUNESERR_SYNTAXNOTDEFINED;
			else /* 4 0+ 8+ (this is the power function) */
				return dup_sub(arg[1],arg[2]);
		}
		else if (arg[0]==5)
		{
			if (arg[1]<8 || arg[2]<2 || arg[2]>3)
				return TUNESERR_SYNTAXNOTDEFINED;
			else if (arg[2]==2) /* 5 8+ 2 */
			{
				err=eval(arg[1]);
				if (err<0)
					return err;
				else
					return del(arg[1]);
			}
			else /* 5 8+ 3 */
			{
				err=display(arg[1]);
				if (err<0)
					return err;
				else
					return del(arg[1]);
			}
		}
	}

	return 0;
}

/* These can't be written to, or read from, so it doesn't matter what we
put in them.  The only reason they exist is because printing the directory
starts at 0 instead of 8, so it won't print out 0-7 as being used unless
memory is allocated for them, and the image file starts at 0 for both
reading and writing, so something has to be here.  I'll just fill them up
with 0 for now.

Returns TUNESERR_NOMEMORY for error or 0 for success
*/
short default_primitives(void)
{
	short dirindex, objindex;

	/* Initialize primitive objects */
	for (dirindex=0;dirindex<8;dirindex++)
	{
		/* Allocate memory for primitive objects */
		dir[dirindex]=malloc(sizeof(tunesobj));
		if (dir[dirindex]==NULL)
			return TUNESERR_NOMEMORY;

		/* Set their contents all to zero. */
		for (objindex=0;objindex<OBJSIZE;objindex++)
			dir[dirindex][objindex]=0;
	}

	for (dirindex=8;dirindex<OBJDIRSIZE;dirindex++)
		dir[dirindex]=NULL;

	return 0;
}

/* TUNES exits with exit code 0 if the image was written, or 1 if not.
   Also 2 for out of memory creating primitives */
int main(void)
{
	short err;

	/* Print a banner message */
	puts("TUNES "TUNES_VERSION);

	/* Display a distribution restriction message */
	puts("Freer than the GPL: Public Domain");

	/* Get some objects, one way or another. */
	err=load(IMAGE_NAME);
	if (err<0)
	{
		switch(err)
		{
			case TUNESERR_FILEOPENREAD:
				puts("Error opening image file for reading");
				break;
			case TUNESERR_BADIMAGEFILEHEADER:
				puts("Invalid image file header");
				break;
			case TUNESERR_BADIMAGEFILEVERSION:
				puts("Wrong image file version: Only version 0 supported");
				break;
			case TUNESERR_BADIMAGEFILEOBJDIRSIZE:
				printf("Wrong object directory size: Only %u supported\n",OBJDIRSIZE);
				break;
			case TUNESERR_BADIMAGEFILEOBJSIZE:
				printf("Wrong object size: Only %u supported\n",OBJSIZE);
				break;
			case TUNESERR_NOMEMORY:
				puts("Out of memory reading image file");
				break;
			case TUNESERR_BYTECOUNTMISMATCH:
				puts("Wrong number of bytes read from image file");
				break;
		}

		puts("Image file failed, creating default primitives");

		err=default_primitives();
		if (err==TUNESERR_NOMEMORY)
		{
			puts("Out of memory creating primitive objects");
			exit(2);
		}
	}

	/* Blank space */
	puts("");

	/* Interact with the keyboard */
	input_loop();

	/* If control reaches this point, the user hit EOF in the
           input loop instead of typing '7' to quit, which results in
	   the image file not getting saved. */
	return 1;
}

/* Returns if there is no object created in slot 2, or if EOF is received
from stdin
*/
void input_loop(void)
{
	char inputline[20]; /* should be enough for numbers only */
	short *tty; /* Don't use tunesobj* since C hates bounded ptrs */
	short objindex;
	short result; 
	int a, b, c; /* for sscanf, which doesn't take shorts */

	/* Find the place to store console commands */
	tty=dir[2];
	if (tty==NULL)
	{
		puts("Fatal error: No object in slot 2 to store commands");
		return;
	}

	/* Clear out the tty object */
	for (objindex=0;objindex<OBJSIZE;objindex++)
		tty[objindex]=-1;

	for(;;)
	{
		printf("> ");	/* Prompt */

		/* Input */
		if (fgets(inputline,20,stdin)==NULL)
			return; /* Quit without saving */

		/* Parse */
		a=b=c=-1;
		sscanf(inputline, "%u %u %u", &a, &b, &c);

		tty[0]=(short)a;
		tty[1]=(short)b;
		tty[2]=(short)c;

		debug ("input_loop: got command: %d %d %d\n",
			tty[0],tty[1],tty[2]);

		/* Evaluate */
		result=eval(2);

		/* Recover from whatever just happened */
		switch(result)
		{
			case TUNESERR_DOESNOTEXIST:
				puts("Object does not exist");
				break;
			case TUNESERR_SYNTAXNOTDEFINED:
				puts("Syntax not defined");
				break;
			case TUNESERR_ACCESSDENIED:
				puts("Cannot delete primitive");
				break;
			case TUNESERR_DIRFULL:
				printf("Object directory full (%u objects)\n",
					OBJDIRSIZE);
				break;
			case TUNESERR_NOMEMORY:
				puts("Out of memory");
				break;
			case TUNESERR_BYTECOUNTMISMATCH:
				puts("Wrong byte count writing to image file");
				break;
			case TUNESERR_FILEOPENWRITE:
				puts("Error opening image file for writing");
				break;
			default:
				printf("Operation returned result: %d\n",
					result);
				break;
		}
	}
}
